/*
   Copyright (C) 2007 - 2009 by Roberto García Carvajal <roberpot@gmail.com>
   Part of libglsws (http://magiccards.wordpress.com)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3
   or at your option any later version.
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.

   See the COPYING file for more details.
*/

#include <fstream>
#include <string>
#include <sstream>
#include "c_lexico.h"

using namespace std;


/*******************************************************************************
*
* Función eschar.
* Recibe un carácter.
* Devuelve true si el carácter es a-z, A-Z o _.
* En caso contrario devuelve false.
*
*******************************************************************************/
bool eschar(GLubyte c)
{
 if((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') || c == '_')
  return true;
 return false;
}

/*******************************************************************************
*
* Función esint.
* Recibe un carácter.
* Devuelve true si es 0-9
* Devuelve false en caso contrario.
*
*******************************************************************************/
bool esint(GLubyte c)
{
 if(c >= '0' && c <= '9')
  return true;
 return false;
}

/*******************************************************************************
*
* Constructor c_lexico.
* Recibe un nombre de fichero.
* Abre el fichero para lectura e inicializa el _token a L_FIN.
*
*******************************************************************************/
c_lexico::c_lexico(const GLbyte * f)
{
 _fichero.open((const char *)f);
 _token = L_FIN;


 // Introduzco las cadenas para el mapeado de cadenas a tokens.
 _mapa_cadenas["ventana"] = L_VENTANA;
 _mapa_cadenas["v_id"] = L_ID;
 _mapa_cadenas["v_x"] = L_X;
 _mapa_cadenas["v_y"] = L_Y;
 _mapa_cadenas["v_img"] = L_IMG;
 _mapa_cadenas["v_ancho"] = L_ANCHO;
 _mapa_cadenas["v_alto"] = L_ALTO;
 _mapa_cadenas["v_visible"] = L_VISIBLE;
 _mapa_cadenas["v_movil"] = L_MOVIL;
 _mapa_cadenas["v_color"] = L_COLOR;
 _mapa_cadenas["v_bcolor"] = L_BCOLOR;
 _mapa_cadenas["v_tcolor"] = L_TCOLOR;
 _mapa_cadenas["v_texto"] = L_TEXTO;
 _mapa_cadenas["v_factortexto"] = L_FACTORTEXTO;
 _mapa_cadenas["v_grosortexto"] = L_GROSORTEXTO;
 _mapa_cadenas["true"] = L_TRUE;
 _mapa_cadenas["false"] = L_FALSE;
 _mapa_cadenas["objeto"] = L_OBJETO;
 _mapa_cadenas["o_boton"] = L_OBOTON;
 _mapa_cadenas["o_imagen"] = L_OIMAGEN;
 _mapa_cadenas["o_texto"] = L_OTEXTO;
 _mapa_cadenas["o_casillero"] = L_OCASILLERO;
 _mapa_cadenas["o_textoeditable"] = L_OTEXTOEDITABLE;
 _mapa_cadenas["o_selector"] = L_OSELECTOR;
 _mapa_cadenas["o_fselector"] = L_OSELECTORFICHEROS;
 _mapa_cadenas["onclick"] = L_ECLICK;
 _mapa_cadenas["onblur"] = L_EBLUR;
 _mapa_cadenas["retorno"] = L_RETORNO;
 _mapa_cadenas["parametros"] = L_PARAMETROS;

 // Introduzco las cadenas para saber que carácteres debo interpretar.
 _conjunto_caracteres.insert('{');
 _conjunto_caracteres.insert('}');
 _conjunto_caracteres.insert('[');
 _conjunto_caracteres.insert(']');
 _conjunto_caracteres.insert('(');
 _conjunto_caracteres.insert(')');
 _conjunto_caracteres.insert(',');
 _conjunto_caracteres.insert('=');
 _conjunto_caracteres.insert('.');
}

/*******************************************************************************
*
* Método _leer_cadena()
* Lee una cadena y prepara su token.
*
*******************************************************************************/

bool c_lexico::_leer_cadena()
{
 string cadena;
 GLubyte c;

 c = _fichero.get();
 while(eschar(c))
 {
  cadena.push_back(c);
  c = _fichero.get();
 }
 _fichero.unget();

 // Compruebo que sea uno de los lexemas esperados...
 if(_mapa_cadenas.count(cadena))
 {
  _token = _mapa_cadenas[cadena];
  return false;
 }

 _token = L_ERROR;
 return false;
}

/*******************************************************************************
*
* Método _leer_numero()
* Lee un número y prepara su token y su lexema.
*
*******************************************************************************/

bool c_lexico::_leer_numero()
{
 bool entero = true;
 string cadena;
 GLubyte c;

 c = _fichero.get();
 while(esint(c) || c == '.')
 {
  if(c == '.' && entero)
  {
   cadena.push_back(c);
   entero = false;
   c = _fichero.get();
  }
  else if(c == '.' && !entero)
  {
   c = ' ';
  }
  else
  {
   cadena.push_back(c);
   c = _fichero.get();
  }
 }
 _fichero.unget();

 if(entero)
 {
  istringstream buffer(cadena);
  buffer >> _i_lexema;
  _token = L_ENTERO;
  return false;
 }
 else
 {
  istringstream buffer(cadena);
  buffer >> _f_lexema;
  _token = L_FLOTANTE;
  return false;
 }
}

bool c_lexico::_leer_dato_cadena()
{
 GLubyte c;
 string cadena;

 _fichero.get();
 c = _fichero.get();

 while(c != '"' && c != '\000')
 {
  cadena.push_back(c);
  c = _fichero.get();
 }
 if(c != '\000')
 {
  _s_lexema = cadena;
  _token = L_CADENA;
 }
 else
  _token = L_ERROR;

 return false;
}

/*******************************************************************************
*
* Método leer()
* Lee el siguiente token del fichero.
*
*******************************************************************************/
void c_lexico::leer()
{
 // Si hay algo en al _cinta, cojo de la _cinta.
 if(_cinta.size())
 {
  _token = _cinta.front();
  _cinta.pop_front();
 } // Si no...
 else
 {
  bool continuar = true;
  GLubyte c;

  // Si el fichero no está al inal.
  _token = L_FIN;
  if(_fichero.eof())
   return;
  do
  {
   // Cojo un carácter.
   c = _fichero.peek();
   // Si es una letra o _...
   if(eschar(c))
   {
    continuar = _leer_cadena();
   }
   else if(esint(c))
   {
    continuar = _leer_numero();
   }
   else if(c == '"')
   {
    continuar = _leer_dato_cadena();
   }
   else if(_conjunto_caracteres.count(c))
   {
    _token = c;
    continuar = false;
    _fichero.get();
   }
   else if(c == '\000')
   {
    _token = L_FIN;
    continuar = false;
    _fichero.get();
   }
   else if(c == 255)
   {
    _token = L_FIN;
    continuar = false;
    _fichero.get();
   }
   else
   {
    _fichero.get();
   }
  } while(continuar);
 }
}

/*******************************************************************************
*
* Método rebobinar()
* Dejo un token en la cinta.
*
*******************************************************************************/
void c_lexico::rebobinar()
{
 _cinta.push_back(_token);
}

/*******************************************************************************
*
* Método token()
* Devuelvo el token actual.
*
*******************************************************************************/
GLuint c_lexico::token()
{
 return _token;
}
